# -*- coding: UTF-8 -*-
""" Base class for tv checks """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__copyright__ = "Copyright (C) 2014-19, Florian JUNG"
__license__ = "MIT"
__version__ = "0.6.5"
__date__ = "2019-12-05"
# Created: 2014-05-18 04:08

import datetime
from datetime import timedelta
import time
import os
import threading
from collections import OrderedDict
import abc
import shutil

import pytz
from past.builtins import basestring
from future.utils import with_metaclass
import flotils
from flotils.loadable import Loadable, DateTimeEncoder, DateTimeDecoder
from flotils import StartStopable


logger = flotils.get_logger()  # pylint: disable=invalid-name


class TVNextException(Exception):
    """ Base exception for project """
    pass


class TVNextFatalException(Exception):
    """ Exception that can not be recovered from """
    pass


class TVNextNotFoundException(Exception):
    """ Show not found """
    pass


class JSONDecoder(DateTimeDecoder):
    """ Extend DateTimeDecoder with tzinfo field """

    @staticmethod
    def _as_tzinfo(dct):
        if "__pytzinfo__" in dct.keys():
            return pytz.timezone(dct['__pytzinfo__'])
        raise TypeError("Not pytzinfo")

    @staticmethod
    def decode(dct):
        try:
            if isinstance(dct, dict):
                return JSONDecoder._as_tzinfo(dct)
        except:
            pass
        res = super(JSONDecoder, JSONDecoder).decode(dct)
        if isinstance(res, datetime.datetime):
            res = res.replace(tzinfo=pytz.UTC)
        return res


class JSONEncoder(DateTimeEncoder):
    """ Extend DateTimeEncoder with tzinfo field """

    def default(self, obj):
        if isinstance(obj, datetime.tzinfo):
            return obj.zone
        else:
            return super(JSONEncoder, self).default(obj)


class TVNext(with_metaclass(abc.ABCMeta, Loadable, StartStopable)):
#class TVNext(Loadable, StartStopable):
    """
    Abstract checker for new tv episodes
    """
#    __metaclass__ = abc.ABCMeta

    def __init__(self, settings=None):
        """
        Initialise instance

        :param settings: Settings
        :type settings: None | dict
        :rtype: None
        """
        if settings is None:
            settings = {}
        super(TVNext, self).__init__(settings)

        self._shows = OrderedDict(settings.get('shows', {}))
        """ Show information
            :type : dict[unicode, dict[unicode, bool | unicode | \
            datetime.time | datetime.datetime | int | dict[unicode, dict[\
            unicode, unicode | datetime.datetime]]] """
        self._cache_path = settings.get('cache_path', None)
        """ Path to cache directory (default: None)
            :type _cache_path: None | str """
        self._cache_file = settings.get('cache_file', "sallie.tmp.json")
        """ Cache information file on shows (default: sallie.tmp.json)
            :type : str """
        self._show_file = settings.get('show_file', None)
        """ File of show names (default: None)
            :type : None | str """

        self._access_interval = settings.get('update_interval', 7)
        """ Refresh data every x days for active shows (default: 1)
            :type : int """
        self._access_interval_hiatus = settings.get(
            'update_interval_hiatus', None
        )
        """ Refresh interval for shows marked as hiatus in days (default: None)
        :type : None | int """
        self._access_interval_inactive = settings.get(
            'update_interval_inactive', None
        )
        """ Refresh interval for shows marked as inactive in days (default: None)
        :type : None | int """
        self._access_interval_aired = settings.get(
            'update_interval_aired', None
        )
        """ Refresh interval for shows after a new episode aired within \
            the last x days (default: None)
            :type : None | int """
        self._access_interval_error = settings.get(
            'update_interval_error', 10
        )
        """ Refresh interval for shows with error loading (min) (default: 10)
        :type : int """

        self._timezone = settings.get('timezone_default', "US/Pacific")
        """ Timezone to be used if not provided for show
            :type : datetime.timezone """

        if isinstance(self._timezone, basestring):
            # If given as name -> convert to timezone with pytz
            self._timezone = pytz.timezone(self._timezone)

        self._error_sleep_time = settings.get('update_retry_delay', 2.0)
        """ Time to sleep after error
            :type : float """
        self._max_retries = settings.get('update_retry_num', 3)
        """ Maximum number of retries on load failure
            :type : int """
        self._should_retry = settings.get('update_retry', True)
        """ Should be retried
            :type : bool """

        if self._cache_path:
            self._cache_path = self.join_path_prefix(self._cache_path)

            if self._cache_file:
                self._cache_file = os.path.join(
                    self._cache_path, self._cache_file
                )

        self._lock_update = threading.RLock()
        """ Lock to prevent concurrent access
            :type : threading.RLock """

    def start(self, blocking=False):
        """
        Start the interface

        :param blocking: Should the call block until stop() is called
            (default: False)
        :type blocking: bool
        :rtype: None
        """
        if self._cache_file and os.path.exists(self._cache_file):
            shows = self._load_json_file(
                self._cache_file, decoder=JSONDecoder
            )
            shows.update(self._shows)
            self._shows.update(shows)
        # Don't care whether cache file exitsts -> generate it
        # else:
        #    raise IOError("File '{}' not found".format(self._showPath))
        if self._show_file:
            self.show_name_file_load()
        super(TVNext, self).start(blocking)

    def stop(self):
        """
        Stop the interface

        :rtype: None
        """
        self.debug("()")
        super(TVNext, self).stop()
        # self.show_save_all()

    def show_name_file_load(self):
        """
        Load file containing show names

        :rtype: None
        :raises IOError: File not found
        :raises IOError: Failed to load
        """
        self._show_file = self.join_path_prefix(self._show_file)

        if os.path.exists(self._show_file):
            show_names = self._load_json_file(
                self._show_file, decoder=JSONDecoder
            )
            """ :type : list[str | (str, str)] """

            for name in show_names:
                timezone = self._timezone
                if isinstance(name, list):
                    name, timezone = name
                if isinstance(timezone, basestring):
                    timezone = pytz.timezone(timezone)
                self._shows.setdefault(name, {})
                self._shows[name]['air_timezone'] = timezone
        else:
            raise IOError("File '{}' not found".format(self._show_file))
        self.debug("Loaded show names from {}".format(self._show_file))

    def show_add(self, show, timezone=None):
        """
        Add new show with default params

        :param show: Name of show
        :type show: unicode
        :param timezone: Timezone to add - None means no timezone
            (Default: None)
        :rtype: None
        """
        if isinstance(timezone, basestring):
            timezone = pytz.timezone(timezone)
        with self._lock_update:
            self._shows.setdefault(show, {})
            if timezone:
                self._shows[show]['air_timezone'] = timezone

    def show_remove(self, show):
        """
        Remove show

        :param show: Name of show
        :type show: unicode
        :rtype: None
        """
        with self._lock_update:
            if show in self._shows:
                del self._shows[show]

    @property
    def shows(self):
        """
        Show information

        :return: Shows being watched watching
        :rtype: dict[unicode, dict[unicode, bool | unicode | datetime.time | \
        datetime.datetime | int | dict[unicode, dict[\
        unicode, unicode | datetime.datetime]]]
        """
        return self._shows

    def show_save_all(self, path=None):
        """
        Save shows cache

        :param path: Path to save to (default: None)
            if None -> use cache_file
        :type path: None | unicode
        :rtype: None
        """
        if not path:
            path = self._cache_file
        temp_path = path + "temp.json"
        try:
            self._save_json_file(
                temp_path,
                self._shows,
                pretty=True,
                sort=True,
                encoder=JSONEncoder
            )
            shutil.move(temp_path, path)
            self.info("Saved shows to {}".format(path))
        except:
            self.exception("Failed to save shows to {}".format(path))

    def show_episodes_flatten(self, key, show=None):
        """
        Get a (flat) list of this shows episodes

        :param key: Show key
        :type key: unicode
        :param show: Show information (default: None)
            if None -> use key to retrieve show information
        :type show: dict[unicode, bool | unicode | datetime.time | \
        datetime.datetime | int | dict[unicode, dict[\
        unicode, unicode | datetime.datetime]]
        :return: Episode list
        :rtype: list[dict[unicode, unicode | datetime.datetime]]
        """
        if not show:
            show = self._shows[key]
        eps = show['episodes']
        return [eps[s_nr][e_nr] for s_nr in eps for e_nr in eps[s_nr]]

    def show_update_should(self, key):
        """
        Decide if show needs updating

        :param key: Show key
        :type key: unicode
        :return: Should show be updated
        :rtype: bool
        """
        show = self._shows[key]
        now = pytz.utc.localize(datetime.datetime.utcnow())

        # set defaults
        show.setdefault('active', True)
        show.setdefault('hiatus', False)
        show.setdefault('episodes', {})
        accessed = show.get('accessed', None)
        episodes = self.show_episodes_flatten(key, show)

        diff = None
        """ :type : datetime.timedelta """

        if accessed:
            diff = now - accessed
        else:
            self.info("Never accessed {}".format(key))
            return True
        # Update if errors last time
        if show.get('errors', 0) >= 3 and \
                diff.days >= self._access_interval_error:
            self.info("Error retry {}".format(key))
            # Only retry once
            show['errors'] = max(1, self._max_retries - 1)
            return True

        if self._access_interval_aired is not None:
            # Update after new episode aired (in last x days)
            # Reverse list (newest first)
            for ep in episodes[::-1]:
                aired = ep['aired']
                """ :type : datetime.datetime """
                if not aired:
                    # Skip with no air dates
                    continue
                # TODO: make sure tzinfo is correctly loaded
                if aired.tzinfo is None:
                    self.warning("Unset tzinfo in {} ({})".format(
                        key, ep['id']
                    ))
                    aired = pytz.utc.localize(aired)
                if aired > now:
                    # Skip not aired yet
                    continue
                if (aired - accessed).days >= self._access_interval_aired:
                    # Last access before aired
                    self.debug("Aired {} ago {}".format(now - aired, key))
                    return True
                else:
                    # Out of reach
                    break
        if not show['active']:
            # Update inactive shows every so often in case they were mislabeled
            if diff is not None and self._access_interval_inactive:
                # self.debug("Inactive {} ago {}".format(diff, key))
                return diff.days >= self._access_interval_inactive
            # self.debug("{} inactive".format(key))
            return False
        if show['hiatus']:
            # Update shows on hiatus every so often in case episodes got added
            if diff is not None and self._access_interval_hiatus:
                # self.debug("Hiatus {} ago {}".format(diff, key))
                return diff.days >= self._access_interval_hiatus
            # self.debug("{} hiatus".format(key))
            return False

        # self.debug("Last access {} ago {}".format(diff, key))
        # Use generic access interval
        return diff.days >= self._access_interval

    def show_update(self, key, force_check=False, auto_save=False):
        """
        Update show (and adds it if not already present)

        :param force_check: Force information load (default: False)
        :type force_check: bool
        :param key: Show name
        :type key: unicode
        :param auto_save: Save after update (default: False)
        :type auto_save: bool
        :return: Changed
        :rtype: bool
        """
        if not (force_check or self.show_update_should(key)):
            # Show information already up to date
            return False
        changed = False
        retry = True
        error_sleep = self._error_sleep_time
        self.shows[key].setdefault('errors', 0)

        while retry \
                and self.shows[key]['errors'] < self._max_retries:
            retry = False

            # TODO: move into implementation (Backoff exception?) - fatal
            try:
                changed = self._show_update(key)
            except TVNextNotFoundException:
                # Show not found
                self.shows[key]['errors'] = self._max_retries
                self.shows[key]['accessed'] = pytz.utc.localize(
                    datetime.datetime.utcnow()
                )
                changed = True
                break
            except Exception as e:
                if "connection reset by peer" in "{}".format(e).lower():
                    # Connection reset by peer -> exponential backoff
                    self.warning(
                        "Connection reset by peer (Sleeping {})".format(
                            error_sleep
                        )
                    )
                    time.sleep(error_sleep)
                    error_sleep *= 2.0
                else:
                    self.exception("Failed to load {}".format(key))
                self.shows[key]['errors'] += 1
                retry = self._should_retry
                changed = True

        if auto_save and changed:
            self.show_save_all()
        return changed

    @abc.abstractmethod
    def _show_update(self, key):
        """
        Update show (and adds it if not already present)

        Overwrite to implement

        :param key: Show name
        :type key: unicode
        :return: Changed
        :rtype: bool
        """
        raise NotImplementedError

    def show_update_all(self, force_check=False, auto_save=False):
        """
        Update information for all shows

        :param force_check: Force information load (default: False)
        :type force_check: bool
        :param auto_save: Save after each update (default: False)
        :type auto_save: bool
        :rtype: None
        """
        shows = self._shows
        changed = False
        for key in sorted(shows):
            changed = changed or self.show_update(key, force_check, auto_save)
        if not auto_save and changed:
            self.show_save_all()

    def check(
            self, key, force_check=False, delta_min=None, delta_max=None,
            auto_save=False
    ):
        """
        Check a single show for updates in a period
        (adds it if not already present and updates info)

        :param key: Show to check
        :type key: unicode
        :param force_check:  Force information load (default: False)
        :type force_check: bool
        :param delta_min: Time in the past (either date or delta from now)
        :type delta_min: datetime.date | datetime.timedelta | int
        :param delta_max: Time in the future (either date or delta from now)
        :type delta_max: datetime.date | datetime.timedelta | int
        :param auto_save: Save after each update (default: False)
        :type auto_save: bool
        :return: Show, episode pair matching the query
        """
        if not self.is_running:
            self.warning("Not running")
        # Delta as ints = days
        if isinstance(delta_min, int):
            delta_min = timedelta(days=delta_min)
        if isinstance(delta_max, int):
            delta_max = timedelta(days=delta_max)
        # Delta defaults
        if delta_min is None:
            delta_min = timedelta()
        if delta_max is None:
            delta_max = timedelta()

        now = pytz.utc.localize(datetime.datetime.utcnow()).date()
        if isinstance(delta_min, datetime.date):
            # Delta as date
            day_min = delta_min
        else:
            # Delta as timedelta
            day_min = now - delta_min
        if isinstance(delta_max, datetime.date):
            # Delta as date
            day_max = delta_max
        else:
            # Delta as timedelta
            day_max = now + delta_max

        # Wrong order
        if day_min > day_max:
            day_min, day_max = day_max, day_min

        with self._lock_update:
            if key not in self.shows:
                # Add missing
                self.show_add(key)
            # Update info
            changed = self.show_update(key, force_check, auto_save)

            # self.info("Searching between {} and {}".format(day_min, day_max))
            show = self.shows[key]
            episodes = self.show_episodes_flatten(key, show)
            results = []

            # More likely that recent epsiode selected -> backwards
            for ep in episodes[::-1]:
                d = ep['aired']
                if not d:
                    # self.debug(u"No aired: {}.{}".format(key, ep['id']))
                    continue
                d = d.date()
                # if d < day_min:
                #    break
                if d > day_max or d < day_min:
                    # Out of range
                    continue
                results.append((key, ep.copy()))
            return results

    def check_all(
            self, keys=None, delta_min=None, delta_max=None,
            force_check=False, auto_save=False
    ):
        if not self._is_running:
            self.warning("Not running")
        results = []
        with self._lock_update:
            if keys is None:
                keys = sorted(list(self.shows))
                self.show_update_all(force_check, auto_save)
                # All updates allready done
                force_check = False
                auto_save = False
            for key in keys:
                results.extend(
                    self.check(
                        key, force_check=force_check, auto_save=auto_save,
                        delta_min=delta_min, delta_max=delta_max
                    )
                )
        return results



# -*- coding: UTF-8 -*-
""" TVNext implementation for tvdb """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__copyright__ = "Copyright (C) 2016-19, Florian JUNG"
__license__ = "MIT"
__version__ = "0.2.7"
__date__ = "2019-12-05"
# Created: 2015-04-29 19:15

import datetime

import flotils
import dateutil
import dateutil.parser
import pytz
from past.builtins import basestring
import tvdb_api
from tvdb_api import BaseUI
from tvdb_api import tvdb_attributenotfound, tvdb_episodenotfound, \
    tvdb_seasonnotfound, tvdb_shownotfound, tvdb_userabort

from .tv_next import TVNext, TVNextNotFoundException


logger = flotils.get_logger()


class TVNextTVDB(TVNext):
    """
    Check tv shows with tvdb
    """

    class TVDBUI(BaseUI):
        """ Class to select a show based on year """
        # pylint: disable=invalid-name

        def selectSeries(self, allSeries):
            key = self.getSearchKey()
            spl = key.split()
            if spl[-1].startswith("(") and spl[-1].endswith(")"):
                year = spl[-1][1:-1]
                for series in allSeries:
                    if series.get('firstaired') and \
                            series['firstaired'].split('-')[0] == year:
                        # Same year -> assume this one
                        return series
            return allSeries[0]

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(TVNextTVDB, self).__init__(settings)

        cache = True
        if self._cache_path:
            cache = self._cache_path
        self.debug("Using cache at {}".format(self._cache_path))
        self.debug("Using cache file {}".format(self._cache_file))
        self._tvdb_cache = cache
        self._tvdb = None
        self._init_tvdb()

    def _init_tvdb(self):
        self.info("Initializing tvdb instance")
        self._tvdb = tvdb_api.Tvdb(
            cache=self._tvdb_cache, custom_ui=self.TVDBUI
        )
        if self._tvdb_cache:
            # Fix tvdb_api bug
            self._tvdb.config['cache_enabled'] = True
        else:
            self._tvdb.config['cache_enabled'] = False

    def _key_error_retry(self, key):
        if self._tvdb is None:
            self._init_tvdb()
        try:
            return self._tvdb[key]
        except KeyError as e:
            self.warning("{}: Key error '{}'".format(key, e))
            self._init_tvdb()
            return self._tvdb[key]

    def _show_update(self, key):
        """
        Update show (and adds it if not already present)

        :param key: Show name
        :type key: str
        :return: Changed (Might not really have changed - but successfull read)
        :rtype: bool
        """
        def _get_search_key(other):  # pylint: disable=unused-argument
            return key

        self.debug("Updating {}..".format(key))
        changed = False
        show = self._shows.setdefault(key, {})
        now = pytz.utc.localize(datetime.datetime.utcnow())
        self.TVDBUI.getSearchKey = _get_search_key

        try:
            tvdb_show = self._key_error_retry(key)
        except tvdb_shownotfound as e:
            year = key.split()[-1]

            if year.startswith("(") and year.endswith(")"):
                # Assume year info
                new_key = key.rstrip(" " + year)
                self.info("Trying {} instead".format(new_key))
                # Try without year info
                try:
                    tvdb_show = self._tvdb[new_key]
                except tvdb_shownotfound as e2:
                    self.error("Show {}: {}".format(new_key, e2))
                    raise TVNextNotFoundException(new_key)
            else:
                self.error("Show {}: {}".format(key, e))
                raise TVNextNotFoundException(key)

        # self.debug("{}".format(tvdb_show))
        # for key, item in tvdb_show.data.items():
        #    self.debug("\n {}: {}".format(key, item))
        # for key, item in tvdb_show.items():
        #    self.debug("\n {}: {}".format(key, item))
        # for item in tvdb_show[0][1].keys():
        #    self.debug("\n     {}: {}".format(item, tvdb_show[0][1][item]))
        show['id_tvdb'] = tvdb_show['id']
        show['id_imdb'] = tvdb_show['imdbId']
        show['name'] = tvdb_show['seriesname']
        show['overview'] = tvdb_show['overview']
        show['status'] = tvdb_show.get('status', "").upper()
        show['active'] = show['status'] != "ENDED"
        show['hiatus'] = False
        show['air_duration'] = tvdb_show['runtime']
        show['air_day'] = tvdb_show['airsDayOfWeek']
        show['air_time'] = tvdb_show['airsTime']
        show.setdefault('air_timezone', self._timezone)

        if isinstance(show['air_timezone'], basestring):
            show['air_timezone'] = pytz.timezone(show['air_timezone'])
        if show['air_time']:
            show['air_time'] = dateutil.parser.parse(show['air_time']).time()

        # Either correct time
        # or last possible second on that day (-> upper bound)
        air_time = show['air_time'] or datetime.time(23, 59, 59)
        episodes = {}

        for season_nr in tvdb_show:
            episodes.setdefault(season_nr, {})
            # episode db id name: id_<unique_name_for_source>
            for episode_nr in tvdb_show[season_nr]:
                # for key, item in tvdb_show[season_nr][episode_nr].items():
                #     self.debug("\nep:  {}: {}".format(key, item))
                episodes[season_nr][episode_nr] = {
                    'id_tvdb': tvdb_show[season_nr][episode_nr]['id'],
                    'id_imdb': tvdb_show[season_nr][episode_nr]['imdbId'],
                    'aired': None,
                    'name': tvdb_show[season_nr][episode_nr]['episodeName'],
                    'id': u"{:02d}x{:02d}".format(season_nr, episode_nr)
                }
                aired = tvdb_show[season_nr][episode_nr]['firstAired']
                if aired:
                    # Use datetime instead of date for tz info localization
                    try:
                        aired = dateutil.parser.parse(aired)
                    except:
                        if aired == "0000-00-00":
                            self.error("Invalid aired date: {}-{}x{}".format(
                                key, season_nr, episode_nr
                            ))
                        else:
                            self.exception("Failed to parse date: {}".format(
                                aired
                            ))
                        aired = None
                if aired and air_time:
                    aired = datetime.datetime.combine(
                        aired.date(), air_time
                    )
                    # Set origin tz
                    aired = show['air_timezone'].localize(aired)
                    # Convert to utc
                    aired = aired.astimezone(pytz.utc)
                episodes[season_nr][episode_nr]['aired'] = aired

        show['episodes'] = episodes
        show['accessed'] = now
        show['errors'] = 0
        # Might not really have changed - but successfull read
        return True



# -*- coding: UTF-8 -*-
"""
TV information related tasks
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "the01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2014-19, Florian JUNG"
__license__ = "MIT"
__version__ = "0.6.6"
__date__ = "2019-12-05"
# Created: 2015-02-26 03:39

import logging

from .tv_next import TVNext
from .next_tvdb import TVNextTVDB
from .cli_main import main

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
__all__ = [
    "tv_next", "next_tvdb", "cli_main",
    "TVNext", "TVNextTVDB", "main",
]



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2016-19, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.3"
__date__ = "2019-04-15"
# Created: 2016-05-12 18:04

import argparse
import datetime
import logging.config

import pytz
import flotils
from flotils.logable import default_logging_config

from . import TVNextTVDB


logger = flotils.get_logger()


def shows_format(shows_list):
    """
    Format a list of shows

    :param shows_list: Show list
    :type shows_list: list
    :return: List of formatted text
    :rtype: list[unicode]
    """
    omit_date = True
    date = None
    res = []
    now = pytz.utc.localize(datetime.datetime.utcnow()).date()

    for show, info in shows_list:
        if not info:
            continue
        omit_date &= bool(info.get('aired'))
        if not omit_date:
            break
        if not date:
            date = info['aired'].date()
            continue
        omit_date &= date == info['aired'].date()
    for show, info in shows_list:
        if not info:
            continue
        txt = u""
        name = info.get('name', "")
        nr = info['id']
        if info.get('aired'):
            if omit_date:
                txt += u"{}: ".format(info['aired'].strftime("%H:%M"))
            else:
                dfmt = info['aired'].strftime("%Y-%m-%d %H:%M")
                d = (now - info['aired'].date()).days
                txt += u"{} ({}): ".format(
                    dfmt,
                    u"{}d".format(abs(d))
                    if d < 0 else
                    "today" if d == 0 else u"{}d ago".format(d)
                )
        txt += u"{} ({}".format(show.upper(), nr)

        if name:
            txt += u" - {}".format(name)
        txt += u")"
        res.append(txt)
    return res


def to_string(evs):
    if isinstance(evs, list):
        msg = ""
        last = None
        for ev in sorted(evs, key=lambda x: x[1].get('date')):
            d = ev[1].get('date')
            if d and last != d.date():
                msg += u"{}{}{}".format(
                    "",
                    "\n" + d.strftime("%Y-%m-%d (%a)") + "\n" + "="*10,
                    "\n"
                )
                last = d.date()
            msg += u"  {}{}{}".format(
                "",
                to_string(ev),
                "\n"
            )

        msg += ""
        return msg

    show, info = evs
    name = info.get('name', "")
    date = info.get('date')
    if date:
        date = date.strftime("%Y-%m-%d") + " "
    else:
        date = ""
    #date = ""
    nr = info['episode']
    feedTxt = u"{}{} ({}".format(date, show, nr)

    if name:
        feedTxt += u" - {}".format(name)

    feedTxt += ")"
    return feedTxt


def shows_check(s, autosave, forcecheck, min_delta, max_delta):
    """
    Check for new shows

    :param s: Sallie instance
    :type s: sallie.TVNext
    :param autosave: Save after every request
    :type autosave: bool
    :param forcecheck: Force reload
    :type forcecheck: bool
    :param min_delta: Delta in days into the past
    :type min_delta: None | int
    :param max_delta: Delta in days into the future
    :type max_delta: None | int
    :rtype: None
    """
    if min_delta is not None:
        min_delta = datetime.timedelta(days=min_delta)
    if max_delta is not None:
        max_delta = datetime.timedelta(days=max_delta)
    res = s.check_all(
        delta_max=max_delta,
        delta_min=min_delta,
        auto_save=autosave,
        force_check=forcecheck
    )

    for txt in sorted(shows_format(res)):
        logger.info(txt)
    # logging.info(u"\n{}".format(toString(res)))


def shows_list(
        s, select, sort_by="", reverse=False, list_all=False,
        min_delta=None, max_delta=None):
    """

    :param s: Sallie instance
    :type s: sallie.TVNext
    :param select:
    :type select:
    :param sort_by: Sorts separated by ; (values: date, state)
    :type sort_by: str
    :param reverse: Reverse order (default: False)
    :type reverse: bool
    :param list_all: List all infos (default: False)
    :type list_all: bool
    :param min_delta: Delta in days into the past
    :type min_delta: None | int
    :param max_delta: Delta in days into the future
    :type max_delta: None | int
    :rtype: None
    """
    now = pytz.UTC.localize(datetime.datetime.utcnow())

    def get_state(show):
        if not show:
            return None
        active = show['active']
        hiatus = show['hiatus']
        if not active:
            state = "inactive"
        elif hiatus:
            state = "hiatus"
        else:
            state = "active"
        return state

    def get_aired(ep):
        return ep.get('aired') or datetime.datetime(1, 1, 1, tzinfo=pytz.UTC)

    def in_range(min_delta, ep, max_delta):
        date = get_aired(ep)
        # None -> dont care
        if min_delta is not None and now > date:
            if (now - date) < min_delta:
                return True
            return False
        if max_delta is not None and now <= date:
            if max_delta < (date - now):
                return True
            return False
        return True

    if min_delta is not None:
        min_delta = datetime.timedelta(days=min_delta)
    if max_delta is not None:
        max_delta = datetime.timedelta(days=max_delta)

    shows = s.shows
    sort_by = sort_by.split(";")
    if list_all:
        l = []
        if "state":
            res = {
                "active": [],
                "hiatus": [],
                "inactive": []
            }
            for key in sorted(shows):
                # logger.debug(shows[key])
                state = get_state(shows[key])
                if not state:
                    continue
                res[state].append(key)
            l.extend(res['active'])
            l.extend(res['hiatus'])
            l.extend(res['inactive'])
        if reverse:
            l = reversed(l)
        for key in l:
            state = get_state(shows[key])
            logger.info(u"{} ({})".format(key, state.upper()))
        return

    if select != "all":
        if select not in shows:
            logger.error(u"Show {} not found".format(select))
            return
        keys = [select]
    else:
        keys = shows.keys()
    unformatted = []
    """ :type: list[(str, object)] """

    for key in keys:
        eps = sorted(
            filter(
                lambda x: in_range(min_delta, x, max_delta),
                s.show_episodes_flatten(key)
            ),
            key=get_aired
        )
        unformatted.extend([(key, ep) for ep in eps])

    formatted = shows_format(unformatted)
    if "date" in sort_by:
        formatted = sorted(formatted, reverse=True)
    if reverse:
        formatted = reversed(formatted)
    for txt in formatted:
        logger.info(txt)


def shows_next(s, select, sort_by="", reverse=False, max_delta=None):
    """
    Show next aired episode of show(s)

    :param s: Sallie instance
    :type s: sallie.TVNext
    :param sort_by: Sorts separated by ; (values: date, state)
    :type sort_by: str
    :param reverse: Reverse order (default: False)
    :type reverse: bool
    :param max_delta: Delta in days into the future
    :type max_delta: None | int
    :return: List of episodes; None on failure
    :rtype: None | list[unicode]
    """
    now = pytz.UTC.localize(datetime.datetime.utcnow())

    def get_aired(ep):
        return ep.get('aired') or datetime.datetime(1, 1, 1, tzinfo=pytz.UTC)

    def in_range(min_delta, ep, max_delta):
        date = get_aired(ep)
        res = True
        # None -> dont care
        if min_delta is not None and now > date:
            return (now - date) < min_delta
        if max_delta is not None and now <= date:
            return max_delta > (date - now)
        return True

    min_delta = datetime.timedelta(days=0)
    if max_delta is not None:
        max_delta = datetime.timedelta(days=max_delta)

    shows = s.shows

    if select != "all":
        if select not in shows:
            logger.error(u"Show {} not found".format(select))
            return
        keys = [select]
    else:
        keys = shows.keys()
    unformatted = []
    """ :type: list[(str, object)] """

    for key in keys:
        eps = sorted(
            filter(
                lambda x: in_range(min_delta, x, max_delta),
                s.show_episodes_flatten(key)
            ),
            key=get_aired
        )
        if eps:
            unformatted.append((key, eps[0]))

    formatted = shows_format(unformatted)
    if "date" in sort_by:
        formatted = sorted(formatted, reverse=True)
    if reverse:
        formatted = reversed(formatted)
    return formatted


def setup_parser():
    """
    Create and init argument parser

    :return: Argument parser
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(prog="sallie")
    parser.add_argument(
        "--debug", action="store_true",
        help="Use debug level output"
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )
    parser.add_argument(
        "-s", "--settings", nargs="?", default="sallie.json",
        help="Settings file"
    )
    parser.add_argument(
        "--pre_path", nargs="?", default=None,
        help="Base path to use"
    )
    parser.add_argument(
        "-c", "--check", nargs="?", const=True, default=False,
        help="Check for new shows in given period"
    )
    parser.add_argument(
        "--min", nargs="?", type=int, const=None, default=None,
        help="How many days in the past to check"
    )
    parser.add_argument(
        "--max", nargs="?", type=int, const=1, default=None,
        help="How many days in the future to check"
    )
    parser.add_argument(
        "--force", action="store_true",
        help="Force updates independently from update rules"
    )
    parser.add_argument(
        "--autosave", action="store_true",
        help="Auto save after every show"
    )
    parser.add_argument(
        "--missing", action="store_true",
        help="Which shows have no episodes"
    )
    parser.add_argument(
        "--update", nargs="?", const="all", default=None,
        help="Run a show update; Default: all - check for all shows or "
             "specify a show"
    )
    parser.add_argument(
        "--shows", nargs="?", const=True, default=False,
        help="List all monitored shows"
    )
    parser.add_argument(
        "--list", nargs="?", const="all", default=None,
        help="List episodes; Default: all - check for all shows or "
             "specify a show"
    )
    parser.add_argument(
        "--next", nargs="?", const="all", default=None,
        help="Show next upcoming episode; Default: all - check for all shows "
             "or specify a show"
    )
    parser.add_argument(
        "--sort", nargs="?", type=str, const="date;state", default="date;state",
        help="Order to sort outout; Default: date;state"
    )
    parser.add_argument(
        "--reverse", action="store_false",
        help="Reverse order"
    )
    parser.add_argument(
        "--all", nargs="?", const=True, default=False,
        help="Do to all"
    )

    return parser


def main():
    logging.config.dictConfig(default_logging_config)
    logging.getLogger().setLevel(logging.INFO)

    parser = setup_parser()
    args = parser.parse_args()

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    tv = TVNextTVDB({
        'settings_file': args.settings,
        'path_prefix': args.pre_path
    })

    check = args.check
    tv.start(blocking=False)
    try:
        if args.update:
            check = False
            if args.update == "all":
                tv.show_update_all(
                    force_check=args.force, auto_save=args.autosave
                )
                check = True
            elif args.update not in tv.shows:
                logger.error("Show {} not found".format(args.update))
            else:
                tv.show_update(args.update, auto_save=args.autosave)
        if args.list:
            shows_list(
                tv, args.list, args.sort, args.reverse, args.all,
                args.min, args.max
            )
        elif args.next:
            eps = shows_next(tv, args.next, args.sort, args.reverse, args.max)
            if eps:
                for ep in eps:
                    logger.info("{}".format(ep))
        elif args.shows:
            for key in sorted(tv.shows, reverse=not args.reverse):
                logger.info("{}".format(key))
        elif args.missing:
            for key in sorted(tv.shows, reverse=not args.reverse):
                if len(tv.show_episodes_flatten(key)) == 0:
                    logger.info("Missing episodes for {}".format(key))
        else:
            check = True
        if check:
            shows_check(
                tv,
                autosave=args.autosave, forcecheck=args.force,
                min_delta=args.min, max_delta=args.max
            )
    finally:
        tv.stop()


if __name__ == "__main__":
    main()
